import os
from typing import Union


class CPKException(RuntimeError):

    def __init__(self, msg: str):
        super(RuntimeError, self).__init__(msg)


class NotACPKProjectException(CPKException):

    def __init__(self, path: str):
        super(NotACPKProjectException, self).__init__(
            f"The path '{path}' does not appear to be a CPK project. " +
            (f"The metadata file 'project.cpk' is missing." if os.path.isdir(path) else
             "Path does not exist."))


class InvalidCPKProjectFile(CPKException):

    def __init__(self, path: str, reason: Union[str, None] = None):
        super(InvalidCPKProjectFile, self).__init__(
            f"The path `{path}` contains an invalid CPK project file." + (
                f" Reason: {reason}" if reason else ""
            ))


class CPKProjectSchemaNotSupported(CPKException):

    def __init__(self, schema: str):
        super(CPKProjectSchemaNotSupported, self).__init__(
            f"The project schema used in this project (`{schema}`) is not supported by this "
            f"version of CPK.")


class InvalidCPKTemplate(CPKException):

    def __init__(self, reason: Union[str, None] = None):
        super(InvalidCPKTemplate, self).__init__(
            f"Found an invalid CPK template. Reason: {reason}" if reason else "")


class InvalidCPKTemplateFile(CPKException):

    def __init__(self, path: str, reason: Union[str, None] = None):
        super(InvalidCPKTemplateFile, self).__init__(
            f"The path `{path}` contains an invalid CPK template file." + (
                f" Reason: {reason}" if reason else ""
            ))


class CPKTemplateSchemaNotSupported(CPKException):

    def __init__(self, schema: str):
        super(CPKTemplateSchemaNotSupported, self).__init__(
            f"The template schema used in this project (`{schema}`) is not supported by this "
            f"version of CPK.")


class CPKMissingResourceException(CPKException):

    def __init__(self, resource: str, explanation: str = None):
        exp = "" if explanation is None else f"{explanation}\n"
        super(CPKMissingResourceException, self).__init__(f"{exp}Missing resource: {resource}")


class CPKProjectBuildException(CPKException):

    def __init__(self, reason: Union[str, BaseException]):
        super(CPKProjectBuildException, self).__init__(f"The project failed to build. "
                                                       f"Reason:\n{str(reason)}")


class CPKProjectPushException(CPKException):

    def __init__(self, reason: Union[str, BaseException]):
        super(CPKProjectPushException, self).__init__(f"The project failed to be pushed. "
                                                       f"Reason:\n{str(reason)}")



from ..types import Machine

DEFAULT_TCP_PORT = "2375"
DEFAULT_MACHINE = "unix:///var/run/docker.sock"
DEFAULT_REGISTRY = "docker.io"

DOCKER_INFO = """
Docker Endpoint:
  Machine: {machine}
  Hostname: {Name}
  Operating System: {OperatingSystem}
  Kernel Version: {KernelVersion}
  OSType: {OSType}
  Architecture: {Architecture}
  Total Memory: {MemTotal}
  CPUs: {NCPU}
"""


def transfer_image(origin: Machine, destination: Machine, image, image_size):
    # TODO: re-implement this using in-Python sockets
    # monitor_info = "" if which("pv") else " (install `pv` to see the progress)"
    # cpklogger.info(f'Transferring image "{image}": [{origin}] -> [{destination}]'
    #                f'{monitor_info}...')
    # data_source = ["docker", "-H=%s" % origin, "save", image]
    # data_destination = ["docker", "-H=%s" % destination, "load"]
    # progress_monitor = ["|", "pv", "-cN", "image", "-s", image_size] if which("pv") else []
    # cmd = data_source + progress_monitor + data_destination
    # TODO: re-enable this
    # start_command_in_subprocess(cmd, nostdout=True)
    pass



import argparse
from typing import Optional

from .generic import ProjectVersion
from .null import NullAdapter


class CLIAdapter(NullAdapter):

    def __init__(self, parsed: Optional[argparse.Namespace]):
        super(CLIAdapter, self).__init__('cli')
        self._parsed = parsed
        self.enabled = parsed is not None

    @property
    def version(self) -> ProjectVersion:
        if not self.enabled or not hasattr(self._parsed, 'tag') or self._parsed.tag is None:
            return super(CLIAdapter, self).version
        # ---
        return ProjectVersion(
            tag=self._parsed.tag,
            head=None,
            closest=None,
            sha=None
        ) if hasattr(self._parsed, 'tag') else None



import argparse
import copy
import json
import os
import re
from distutils.dir_util import copy_tree
from typing import Optional

from cpk.cli.commands.info import CLIInfoCommand
from .. import AbstractCLICommand
from ..logger import cpklogger
from ...types import Machine, Arguments

skel_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "..", "skeleton"))
surgery = {
    "name": {
        "title": "Project Name",
        "targets": ["project.cpk"],
        "pattern": r"^[a-zA-Z0-9-_]+$",
        "pattern_human": "an alphanumeric string [a-zA-Z0-9-_]"
    },
    "description": {
        "title": "Project Description",
        "targets": ["project.cpk"],
        "pattern": r"^.*$",
        "pattern_human": "free text"
    },
    "organization": {
        "title": "Owner Username",
        "targets": ["project.cpk"],
        "pattern": r"^[a-zA-Z0-9-_]+$",
        "pattern_human": "an alphanumeric string [a-zA-Z0-9-_]"
    },
    "maintainer": {
        "title": "Owner Full Name",
        "targets": ["project.cpk"],
        "pattern": r"^.*$",
        "pattern_human": "First Last <EMail Address>"
    },
}


class CLICreateCommand(AbstractCLICommand):

    KEY = 'create'

    @staticmethod
    def parser(parent: Optional[argparse.ArgumentParser] = None,
               args: Optional[Arguments] = None) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(parents=[parent])
        parser.add_argument(
            'path',
            default=None,
            help="(optional) Location where to create the new project"
        )
        return parser

    @staticmethod
    def execute(machine: Optional[Machine], parsed: argparse.Namespace) -> bool:
        # handle custom path
        if parsed.path is not None:
            parsed.workdir = os.path.abspath(parsed.path)

        # make sure the path does not exist or it is empty
        if os.path.exists(parsed.workdir) and any(os.scandir(parsed.workdir)):
            cpklogger.error(f"Directory '{parsed.workdir}' is not empty.")
            return False

        # collect info about the new project
        cpklogger.info("Please, provide information about your new project:")
        project_info = {}
        surgery_targets = set()
        print("   |")
        for key in ["name", "description", "organization", "maintainer"]:
            info = surgery[key]
            done = False
            while not done:
                res = input(f"   |\t{info['title']}: ")
                if not re.match(info['pattern'], res):
                    cpklogger.error(f"\tField '{info['title']}' must be {info['pattern_human']}.")
                    continue
                done = True
                project_info[key] = res
            surgery_targets.update(info['targets'])

        # make new project
        os.makedirs(parsed.workdir, exist_ok=True)
        cpklogger.info("New Project workspace: {}".format(parsed.workdir))

        # copy skeleton
        cpklogger.debug(f"Creating empty project...")
        copy_tree(skel_dir, parsed.workdir, dry_run=True)
        copy_tree(skel_dir, parsed.workdir)

        # perform surgery
        for target in surgery_targets:
            target_fpath = os.path.join(parsed.workdir, target)
            cpklogger.debug(f"Performing surgery on {target_fpath}...")
            # read target from disk
            with open(target_fpath, "rt") as fin:
                content = json.load(fin)
            # perform surgery
            new_content = copy.deepcopy(content)
            for k, v in content.items():
                if isinstance(v, str):
                    new_content[k] = v.format(**project_info)
            # write target to disk
            with open(target_fpath, "wt") as fout:
                json.dump(new_content, fout, indent=4, sort_keys=True)
        cpklogger.debug("Surgery completed")

        # show info about the new project
        CLIInfoCommand.execute(None, parsed)
        # TODO: add a link to the online docs on how to get started with the new project
        cpklogger.info(f"Your new project was created in '{parsed.workdir}'.")
        # ---
        return True



import argparse
import re
from dataclasses import dataclass
from typing import Optional, Type

from cpk import cpkconfig
from cpk.cli import AbstractCLICommand, cpklogger
from cpk.machine import SSHMachine
from cpk.types import Machine, Arguments


@dataclass
class MachineTarget:
    pattern: str
    explanation: str
    cls: Type[Machine]


_valid_targets = {
    "ssh": MachineTarget(
        pattern=r"^(ssh\:\/\/)?(?P<user>.*?)@(?P<host>[^:]+?)(?::(?P<port>[0-9]+))?$",
        explanation="user@host[:port]",
        cls=SSHMachine
    ),
}


class CLIMachineCreateCommand(AbstractCLICommand):

    KEY = 'machine create'

    @staticmethod
    def parser(parent: Optional[argparse.ArgumentParser] = None,
               args: Optional[Arguments] = None) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(parents=[parent], add_help=False)
        parser.add_argument(
            'name',
            type=str,
            help="Name to assign to the machine"
        )
        parser.add_argument(
            'target',
            type=str,
            help="Target URI of the machine's endpoint"
        )
        # ---
        return parser

    @staticmethod
    def execute(machine: Machine, parsed: argparse.Namespace) -> bool:
        # validate machine's name
        if not re.match(r"^[a-zA-Z][a-zA-Z0-9\-_.]+[a-zA-Z0-9]$", parsed.name):
            cpklogger.error("Invalid name for a machine. "
                            "Valid characters are: a-z, A-Z, 0-9, -, _\n"
                            "First and last character must be alphanumeric.")
            return False
        # make sure the machine does not exist
        if parsed.name in cpkconfig.machines:
            cpklogger.error("A machine with the same name already exists.")
            return False
        # validate target
        machine = None
        for target_type, target in _valid_targets.items():
            match = re.match(target.pattern, parsed.target)
            if match:
                machine = target.cls(name=parsed.name, **match.groupdict())
                break
        if machine is None:
            cpklogger.error("Invalid target.\nSupported options are:")
            print()
            for target_type, target in _valid_targets.items():
                print(f"\t{target_type}:\t {target.explanation}")
            print()
            return False
        # we have a good machine
        cpklogger.info("Creating machine:")
        print(machine)
        try:
            machine.save(cpklogger)
        except (FileNotFoundError, ValueError) as e:
            cpklogger.error(f"An error occurred while creating the machine. Reason:\n\t{str(e)}")
            machine.remove(cpklogger)
            return False
        except KeyboardInterrupt as e:
            machine.remove(cpklogger)
            raise e
        # ---
        cpklogger.info("Machine successfully created.")
        cpklogger.info("You can now use it with,")
        cpklogger.info(f"\n\t   >  cpk -H {machine.name} ...\n")



import sys
import math


class ProgressBar:

    def __init__(self, scale=1.0, buf=sys.stdout, header="Progress"):
        self._finished = False
        self._buffer = buf
        self._header = header
        self._last_value = -1
        self._scale = max(0.0, min(1.0, scale))
        self._max = int(math.ceil(100 * self._scale))

    def set_header(self, header):
        self._header = header

    def update(self, percentage):
        percentage_int = int(max(0, min(100, percentage)))
        if percentage_int == self._last_value:
            return
        percentage = int(math.ceil(percentage * self._scale))
        if self._finished:
            return
        # compile progress bar
        pbar = f"{self._header}: [" if self._scale > 0.5 else "["
        # progress
        pbar += "=" * percentage
        if percentage < self._max:
            pbar += ">"
        pbar += " " * (self._max - percentage - 1)
        # this ends the progress bar
        pbar += "] {:d}%".format(percentage_int)
        # print
        self._buffer.write(pbar)
        self._buffer.flush()
        # return to start of line
        self._buffer.write("\b" * len(pbar) + "\x1b[2K")
        # end progress bar
        if percentage >= self._max:
            self._buffer.write("Done!\n")
            self._buffer.flush()
            self._finished = True
        self._last_value = percentage_int

    def done(self):
        self.update(100)



import argparse
import logging
import os
import sys

import cpk
from cpk import cpkconfig
from cpk.cli.commands.create import CLICreateCommand
from cpk.exceptions import CPKException

from cpk.cli.logger import cpklogger
from cpk.cli.commands.info import CLIInfoCommand
from cpk.cli.commands.build import CLIBuildCommand
from cpk.cli.commands.run import CLIRunCommand
from cpk.cli.commands.clean import CLICleanCommand
from cpk.cli.commands.push import CLIPushCommand
from cpk.cli.commands.decorate import CLIDecorateCommand
from cpk.cli.commands.machine import CLIMachineCommand
from cpk.cli.commands.endpoint import CLIEndpointCommand
from cpk.utils.machine import get_machine

_supported_commands = {
    'create': CLICreateCommand,
    'info': CLIInfoCommand,
    'build': CLIBuildCommand,
    'run': CLIRunCommand,
    'clean': CLICleanCommand,
    'push': CLIPushCommand,
    'decorate': CLIDecorateCommand,
    'machine': CLIMachineCommand,
    'endpoint': CLIEndpointCommand,
}


def run():
    cpklogger.info(f"CPK - Code Packaging toolKit - v{cpk.__version__}")
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument(
        'command',
        choices=_supported_commands.keys()
    )
    # print help (if needed)
    if len(sys.argv) > 1 and sys.argv[1] in ['-h', '--help']:
        parser.print_help()
        return
    # ---
    # parse `command`
    parsed, remaining = parser.parse_known_args()
    # get command
    command = _supported_commands[parsed.command]
    # let the command parse its arguments
    cmd_parser = command.get_parser(remaining)
    parsed = cmd_parser.parse_args(remaining)
    # sanitize workdir
    parsed.workdir = os.path.abspath(parsed.workdir)
    # enable debug
    if parsed.debug:
        cpklogger.setLevel(logging.DEBUG)
    # get machine
    machine = get_machine(parsed, cpkconfig.machines)
    # avoid commands using `parsed.machine`
    parsed.machine = None
    # execute command
    try:
        with machine:
            command.execute(machine, parsed)
    except CPKException as e:
        cpklogger.error(str(e))
    except KeyboardInterrupt:
        cpklogger.info(f"Operation aborted by the user")


if __name__ == '__main__':
    run()



import argparse

from cpk import CPKProject
from cpk.cli import cpklogger


def check_git_status(project: CPKProject, parsed: argparse.Namespace, must_be_clean: bool = True) \
        -> bool:
    # check if the git HEAD is detached
    if project.is_detached():
        cpklogger.error(
            "The repository HEAD is detached. Create a branch or check one out "
            "before continuing. Aborting."
        )
        return False

    # check if the index is clean
    if project.is_dirty() and must_be_clean:
        cpklogger.warning("Your index is not clean (some files are not committed).")
        cpklogger.warning("If you know what you are doing, use --force (-f) to force.")
        if not parsed.force:
            return False
        cpklogger.warning("Forced!")

    # ---
    return True



from typing import Optional

from .generic import ProjectVersion
from cpk.types import CPKProjectInfo
from .null import NullAdapter


class ProjectFileAdapter(NullAdapter):

    def __init__(self, info: CPKProjectInfo):
        super(ProjectFileAdapter, self).__init__('project.cpk')
        self._info = info
        self.enabled = True

    @property
    def name(self) -> str:
        return self._info.name

    @property
    def registry(self) -> str:
        return self._info.registry

    @property
    def organization(self) -> str:
        return self._info.organization

    @property
    def description(self) -> Optional[str]:
        return self._info.description

    @property
    def version(self) -> ProjectVersion:
        return ProjectVersion(
            tag=self._info.tag or "latest",
            head=self._info.version or None,
            closest=self._info.version or None,
            sha=None
        )

    @property
    def maintainer(self) -> Optional[str]:
        return self._info.maintainer



import argparse
from typing import Optional

import termcolor as tc

from cpk import CPKProject
from .. import AbstractCLICommand
from ..logger import cpklogger
from ...types import Machine, Arguments

PROJECT_INFO = """
{project}
{space} Name: {name}
{space} Tag: {tag}
{space} Version: {version}
{space} Template:
{space}   Name: {template_name}
{space}   Version: {template_version}
{space} Index: {index}
{space} Path: {path}
{space} URL: {url}
{space} Adapters: {adapters}
{end}
"""


class CLIInfoCommand(AbstractCLICommand):

    KEY = 'info'

    @staticmethod
    def parser(parent: Optional[argparse.ArgumentParser] = None,
               args: Optional[Arguments] = None) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(parents=[parent])
        return parser

    @staticmethod
    def execute(machine: Optional[Machine], parsed: argparse.Namespace) -> bool:
        cpklogger.info("Project workspace: {}".format(parsed.workdir))

        # get the project
        project = CPKProject(parsed.workdir)

        # index status
        index = tc.colored("Clean", "green") if project.is_clean() \
            else tc.colored("Dirty", "yellow")

        # show info about project
        info = {
            "project": tc.colored("Project:", "grey", "on_white"),
            "name": project.name,
            "tag": project.version.tag,
            "version": project.version.head or "unreleased",
            "template_name": project.template.name,
            "template_version": project.template.version,
            "index": index,
            "path": project.path,
            "url": project.url or "(none)",
            "adapters": " ".join(project.adapters.keys()),
            "space": tc.colored("  ", "grey", "on_white"),
            "end": tc.colored("________", "grey", "on_white"),
        }
        cpklogger.print(PROJECT_INFO.format(**info))

        # ---
        return True



import argparse
import os
import re
import subprocess
from shutil import which
from typing import Optional, Callable

import cpk
from .. import AbstractCLICommand
from ..logger import cpklogger
from ...types import DockerImageName, Machine, Arguments


class CLIDecorateCommand(AbstractCLICommand):

    KEY = 'decorate'

    DOCKER_IMAGE_REGEX = r"^((?:(?:[a-z0-9]|[a-z0-9][a-z0-9\\-]*[a-z0-9])\\.)*" \
                         r"(?:[a-z0-9]|[a-z0-9][a-z0-9\\-]*[a-z0-9]))(?::([0-9]+)\\/)?" \
                         r"(?:[0-9a-z-]+[/@])(?:([0-9a-z-]+))[/@]?(?:([0-9a-z-]+))?" \
                         r"(?::[a-z0-9\\.-]+)?$"
    EMAIL_ADDRESS_REGEX = r"^[a-z0-9]+[\._]?[a-z0-9]+[@]\w+[.]\w{2,3}$"

    @staticmethod
    def parser(parent: Optional[argparse.ArgumentParser] = None,
               args: Optional[Arguments] = None) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(parents=[parent])
        # ---
        parser.add_argument(
            nargs=1,
            dest="input",
            metavar="INPUT",
            help="The input image",
            type=regex_type(CLIDecorateCommand.DOCKER_IMAGE_REGEX)
        )
        parser.add_argument(
            nargs=1,
            dest="output",
            metavar="OUTPUT",
            help="The output image",
            type=regex_type(CLIDecorateCommand.DOCKER_IMAGE_REGEX)
        )
        parser.add_argument(
            "-m",
            "--maintainer",
            required=True,
            help="The image maintainer's name",
            type=str
        )
        parser.add_argument(
            "-e",
            "--email",
            default=None,
            help="The image maintainer's email address",
            type=regex_type(CLIDecorateCommand.EMAIL_ADDRESS_REGEX)
        )
        return parser

    @staticmethod
    def execute(machine: Machine, parsed: argparse.Namespace) -> bool:
        # pick right value of `arch` given endpoint
        if parsed.arch is None:
            cpklogger.info("Parameter `arch` not given, will resolve it from the endpoint.")
            parsed.arch = machine.get_architecture()
            cpklogger.info(f"Parameter `arch` automatically set to `{parsed.arch}`.")

        # parse `input`
        input_image = DockerImageName.from_image_name(parsed.input[0])
        cpklogger.debug(f"+ Input Image:\n{str(input_image)}")

        # parse `output`
        output_image = DockerImageName.from_image_name(parsed.output[0])

        # append arch to the end of the output image
        if output_image.arch is None:
            output_image.arch = parsed.arch
        cpklogger.debug(f"+ Output Image:\n{str(output_image)}")

        # make sure the `docker` CLI tool is installed
        docker_cli = which('docker')
        if docker_cli is None:
            cpklogger.error("The Docker CLI must be installed for this command to work.")
            return False
        cpklogger.debug(f"Docker CLI found at {docker_cli}")

        # find decorator project inside the `cpk` library
        cpk_dir = os.path.dirname(os.path.abspath(cpk.__file__))
        decorator_dir = os.path.join(cpk_dir, "decorator")
        dockerfile = os.path.join(decorator_dir, "Dockerfile")

        # compile maintainer string
        maintainer = f"{parsed.maintainer} ({parsed.email})" if parsed.email else parsed.maintainer

        # compile command
        cmd = [
            "docker", "build",
            # TODO: machine is not used here
            "-t", output_image.compile(),
            "-f", dockerfile,
            "--build-arg", f"ARCH={parsed.arch}",
            "--build-arg", f"BASE_REGISTRY={input_image.registry.compile(allow_defaults=True)}",
            "--build-arg", f"BASE_ORGANIZATION={input_image.user}",
            "--build-arg", f"BASE_REPOSITORY={input_image.repository}",
            "--build-arg", f"BASE_TAG={input_image.tag}",
            "--build-arg", f"ORGANIZATION={output_image.user}",
            "--build-arg", f"NAME={output_image.repository}",
            "--build-arg", f"MAINTAINER={maintainer}",
            "--build-arg", f"CPK_VERSION={cpk.__version__}",
            decorator_dir
        ]
        cpklogger.info(f"Decorating [{input_image.compile()}] -> [{output_image.compile()}]...")
        cpklogger.debug(f"Running command:\n\t{cmd}")

        # build image
        try:
            subprocess.check_call(cmd)
        except subprocess.SubprocessError as e:
            cpklogger.error(str(e))
            return False

        # success
        cpklogger.info(f"The given image was successfully decorated for CPK.\n"
                       f"Your CPK-compatible image is called\n\n"
                       f"\t\t{output_image.compile()}\n\n"
                       f"You can now use it as a base for your CPK projects and templates.")
        # ---
        return True


def regex_type(pattern: str) -> Callable[[str], str]:
    def _validator(value: str) -> str:
        if not re.compile(pattern).match(value):
            raise argparse.ArgumentTypeError
        return value

    # ---
    return _validator



import os
import logging

from .constants import HEALTH_FILE

# create logger
logging.basicConfig()
logger = logging.getLogger(os.environ.get('CPK_PROJECT_NAME', 'container'))
logger.setLevel(logging.INFO)
if 'DEBUG' in os.environ and os.environ['DEBUG'].lower() in ['true', 'yes', '1']:
    logger.setLevel(logging.DEBUG)


def get_container_health():
    health = "ND"
    if os.path.isfile(HEALTH_FILE):
        try:
            with open(HEALTH_FILE, 'rt') as fin:
                health = fin.read().strip('\n').strip(' ')
        except BaseException:
            logger.warning('An error occurred while trying to fetch the container\'s health.')
    return health


def _set_container_health(new_health):
    health = get_container_health()
    if new_health not in ['healthy', 'unhealthy']:
        logger.warning('Health "{0}" not recognized!'.format(new_health))
        return
    # ---
    if health == new_health:
        return
    logger.info('Updating container health [{0}] -> [{1}]'.format(health, new_health))
    # try to write the health status
    try:
        with open(HEALTH_FILE, 'wt') as fout:
            fout.write(new_health)
    except BaseException:
        logger.warning('An error occurred while trying to set the container\'s health.')


def set_container_healthy():
    _set_container_health('healthy')


def set_container_unhealthy():
    _set_container_health('unhealthy')



import argparse
from typing import Optional

from cpk.utils.misc import human_size

from cpk.utils.docker import DOCKER_INFO

from cpk.cli import AbstractCLICommand, cpklogger
from cpk.types import Machine, Arguments


class CLIEndpointInfoCommand(AbstractCLICommand):

    KEY = 'endpoint info'

    @staticmethod
    def parser(parent: Optional[argparse.ArgumentParser] = None,
               args: Optional[Arguments] = None) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(parents=[parent], add_help=False)
        # ---
        return parser

    @staticmethod
    def execute(machine: Machine, parsed: argparse.Namespace) -> bool:
        # create docker client
        docker = machine.get_client()

        # get info about docker endpoint
        cpklogger.info("Retrieving info about Docker endpoint...")
        epoint = docker.info()
        epoint['machine'] = machine.name
        if "ServerErrors" in epoint:
            cpklogger.error("\n".join(epoint["ServerErrors"]))
            return False
        epoint["MemTotal"] = human_size(epoint["MemTotal"])
        cpklogger.print(DOCKER_INFO.format(**epoint))



import argparse
import os
import shutil
import subprocess
from typing import Optional

from docker.errors import ImageNotFound

from cpk import CPKProject
from cpk.utils.misc import configure_binfmt
from .endpoint import CLIEndpointInfoCommand
from .info import CLIInfoCommand
from .. import AbstractCLICommand
from ..logger import cpklogger
from ...exceptions import NotACPKProjectException
from ...types import CPKFileMappingTrigger, Machine, Arguments
from ...utils.cli import check_git_status

DEFAULT_NETWORK_MODE = "bridge"
SUPPORTED_SUBCOMMANDS = [
    "attach"
]
TRIGGERS = {CPKFileMappingTrigger.DEFAULT, CPKFileMappingTrigger.RUN_MOUNT}


class CLIRunCommand(AbstractCLICommand):

    KEY = 'run'

    @staticmethod
    def parser(parent: Optional[argparse.ArgumentParser] = None,
               args: Optional[Arguments] = None) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(parents=[parent])
        parser.add_argument(
            "-n",
            "--name",
            default=None,
            help="Name of the container"
        )
        parser.add_argument(
            "-c",
            "--cmd",
            default=None,
            help="Command to run in the Docker container"
        )
        parser.add_argument(
            "--pull",
            default=False,
            action="store_true",
            help="Whether to pull the latest image available",
        )
        parser.add_argument(
            "--force-pull",
            default=False,
            action="store_true",
            help="Whether to force pull the image of the project",
        )
        # parser.add_argument(
        #     "--plain",
        #     default=False,
        #     action="store_true",
        #     help="Whether to run the image without default module configuration",
        # )
        parser.add_argument(
            "--no-multiarch",
            default=False,
            action="store_true",
            help="Whether to disable multiarch support (based on bin_fmt)",
        )
        parser.add_argument(
            "-M",
            "--mount",
            default=False,
            const=True,
            action="store",
            nargs="?",
            type=str,
            help="Whether to mount the current project into the container. "
                 "Pass a comma-separated list of paths to mount multiple projects",
        )
        parser.add_argument(
            "--keep",
            default=False,
            action="store_true",
            help="Whether to NOT remove the container once stopped"
        )
        parser.add_argument(
            "-L",
            "--launcher",
            default=None,
            help="Launcher to invoke inside the container",
        )
        parser.add_argument(
            "-A",
            "--argument",
            dest="arguments",
            default=[],
            action="append",
            help="Arguments for the container command",
        )
        parser.add_argument(
            "--runtime",
            default="docker",
            type=str,
            help="Docker runtime to use to run the container"
        )
        parser.add_argument(
            "-X",
            dest="use_x_docker",
            default=False,
            action="store_true",
            help="Use x-docker as runtime",
        )
        parser.add_argument(
            "-s",
            "--sync",
            default=False,
            action="store_true",
            help="Sync code from local project to remote"
        )
        parser.add_argument(
            "--net", "--network_mode",
            dest="network_mode",
            default=DEFAULT_NETWORK_MODE,
            type=str,
            help="Docker network mode"
        )
        parser.add_argument(
            "-d",
            "--detach",
            default=False,
            action="store_true",
            help="Detach from the container and let it run"
        )
        parser.add_argument(
            "--tag",
            default=None,
            type=str,
            help="Custom tag"
        )
        parser.add_argument(
            "docker_args",
            nargs="*",
            default=[]
        )
        # parser.add_argument(
        #     "subcommand",
        #     nargs="?",
        #     default=None,
        #     choices=SUPPORTED_SUBCOMMANDS,
        #     help="(Optional) Subcommand to execute"
        # )
        return parser

    @staticmethod
    def execute(machine: Machine, parsed: argparse.Namespace) -> bool:
        # get project
        project = CPKProject(parsed.workdir, parsed=parsed)

        # show info about project
        CLIInfoCommand.execute(machine, parsed)

        # check git workspace status
        mount_source = parsed.mount is True or isinstance(parsed.mount, str)
        proceed = check_git_status(project, parsed, must_be_clean=mount_source)
        if not proceed:
            return False

        # get info about docker endpoint
        CLIEndpointInfoCommand.execute(machine, parsed)

        # pick right value of `arch` given endpoint
        machine_arch = machine.get_architecture()
        if parsed.arch is None:
            cpklogger.info("Parameter `arch` not given, will resolve it from the endpoint.")
            parsed.arch = machine_arch
            cpklogger.info(f"Parameter `arch` automatically set to `{parsed.arch}`.")

        # create docker client
        docker = machine.get_client()

        # create defaults
        image = project.image(parsed.arch)
        parsed.name = parsed.name or f"cpk-run-{project.name}"

        # subcommand "attach"
        # TODO: this will not work with Machine created from env, the host will be None
        # if parsed.subcommand == "attach":
        #     cpklogger.info(f"Attempting to attach to container '{parsed.name}'...")
        #     # run
        #     _run_cmd(
        #         [
        #             "docker",
        #             "-H=%s" % machine.host,
        #             "exec",
        #             "-it",
        #             parsed.name,
        #             "/entrypoint.sh",
        #             "bash",
        #         ],
        #         suppress_errors=True,
        #     )
        #     return True

        # print info about multiarch
        cpklogger.info("Running an image for {} on {}.".format(parsed.arch, machine_arch))
        # - register bin_fmt in the target machine (if needed)
        if not parsed.no_multiarch:
            configure_binfmt(machine_arch, parsed.arch, docker, cpklogger)

        # ---
        # TODO: this is wrong, x-docker should either be a Python package that gives us the
        #  configuration for X-Forwarding on THIS computer, or a proper docker-runtime that
        #  we can pass to the Python SDK for Docker
        # x-docker runtime
        if parsed.use_x_docker:
            command_dir = os.path.dirname(os.path.abspath(__file__))
            parsed.runtime = os.path.join(command_dir, "x-docker")
        # check runtime
        if shutil.which(parsed.runtime) is None:
            raise ValueError('Docker runtime binary "{}" not found!'.format(parsed.runtime))
        # ---

        # create the module configuration
        module_configuration_args = [
            f"--net={parsed.network_mode}"
        ]

        # mount code
        mount_option = []
        # mount source code (if requested)
        if mount_source:
            # (always) mount current project
            projects_to_mount = [parsed.workdir] if parsed.mount is True else []
            # mount secondary projects
            if isinstance(parsed.mount, str):
                projects_to_mount.extend(
                    [os.path.join(os.getcwd(), p.strip()) for p in parsed.mount.split(",")]
                )
            # create mount points definitions
            for project_path in projects_to_mount:
                # make sure that the project exists
                if not os.path.isdir(project_path):
                    cpklogger.error('The path "{:s}" is not a CPK project'.format(project_path))
                    return False
                # get project info
                try:
                    proj = CPKProject(project_path)
                except NotACPKProjectException:
                    cpklogger.error(f"The path '{project_path}' does not contain a CPK project.")
                    return False
                # iterate over list of mappings
                for mapping in proj.mappings:
                    if TRIGGERS.intersection(set(mapping.triggers)):
                        mpoint_source = mapping.source if os.path.isabs(mapping.source) else \
                            os.path.join(project_path, mapping.source)
                        mpoint_destination = mapping.destination
                        # compile mounpoints
                        mount_option += [
                            "-v", "{:s}:{:s}".format(mpoint_source, mpoint_destination)
                        ]

        # pulling image (if requested)
        if parsed.pull or parsed.force_pull:
            present = False
            try:
                docker.images.get(image)
                present = True
            except ImageNotFound:
                pass
            if present and not parsed.force_pull:
                cpklogger.info("Found an image with the same name. Use --force-pull to pull again")
            else:
                cpklogger.info('Pulling image "%s"...' % image)
                machine.pull_image(image)

        # cmd option
        if parsed.cmd and parsed.launcher:
            raise ValueError("You cannot use the option --launcher together with --cmd.")

        # environment
        environment = []
        if parsed.launcher:
            environment.extend(["-e", f"CPK_LAUNCHER={parsed.launcher}"])

        cmd_option = [] if not parsed.cmd else [parsed.cmd]
        cmd_arguments = (
            [] if not parsed.arguments else ["--"] + list(
                map(lambda s: "--%s" % s, parsed.arguments))
        )

        # endpoint arguments
        docker_epoint_args = []
        # TODO: this should be using the Python SDK for Docker, not Docker CLI
        if machine.base_url is not None:
            docker_epoint_args += ["-H", machine.base_url]

        # docker arguments
        if not parsed.docker_args:
            parsed.docker_args = []
        if (not parsed.keep) or (not parsed.detach):
            parsed.docker_args += ["--rm"]
        if parsed.detach:
            parsed.docker_args += ["-d"]

        # add container name to docker args
        parsed.docker_args += ["--name", parsed.name]
        # escape spaces in arguments
        parsed.docker_args = [a.replace(" ", "\\ ") for a in parsed.docker_args]
        # run
        exitcode = _run_cmd(
            [parsed.runtime]
            + docker_epoint_args
            + ["run", "-it"]
            + module_configuration_args
            + environment
            + parsed.docker_args
            + mount_option
            + [image]
            + cmd_option
            + cmd_arguments,
            suppress_errors=True,
            return_exitcode=True
        )
        if parsed.detach:
            cpklogger.info("Your container is running in detached mode!")
        else:
            cpklogger.debug(f"Command exited with exit code [{exitcode}].")


def _run_cmd(cmd, get_output=False, print_output=False, suppress_errors=False, shell=False,
             return_exitcode=False):
    if shell and isinstance(cmd, (list, tuple)):
        cmd = " ".join([str(s) for s in cmd])
    cpklogger.debug("$ %s" % cmd)
    if get_output:
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=shell)
        proc.wait()
        if proc.returncode != 0:
            if not suppress_errors:
                msg = "The command {} returned exit code {}".format(cmd, proc.returncode)
                cpklogger.error(msg)
                raise RuntimeError(msg)
        out = proc.stdout.read().decode("utf-8").rstrip()
        if print_output:
            print(out)
        return out
    else:
        if return_exitcode:
            res = subprocess.run(cmd, shell=shell)
            return res.returncode
        else:
            try:
                subprocess.check_call(cmd, shell=shell)
            except subprocess.CalledProcessError as e:
                if not suppress_errors:
                    raise e



import logging
import platform
import sys
from functools import partial
from logging import Logger, StreamHandler, Formatter, getLogger

from cpk.constants import DEBUG


# create logger
logging.basicConfig()
cpklogger = getLogger('cpk')
cpklogger.setLevel(logging.INFO)


def setup_logging_format(fmt: str):
    logging.basicConfig(format=fmt)
    # noinspection PyUnresolvedReferences
    root = Logger.root
    if root.handlers:
        for handler in root.handlers:
            if isinstance(handler, StreamHandler):
                formatter = Formatter(fmt)
                handler.setFormatter(formatter)
    else:
        logging.basicConfig(format=fmt)


def add_coloring_to_emit_ansi(fn, indent: int = 0):

    # add methods we need to the class
    def new(*args):
        levelno = args[1].levelno
        if levelno >= 50:
            color = "\x1b[31m"  # red
        elif levelno >= 40:
            color = "\x1b[31m"  # red
        elif levelno >= 30:
            color = "\x1b[33m"  # yellow
        elif levelno >= 20:
            color = "\x1b[32m"  # green
        elif levelno >= 10:
            color = "\x1b[35m"  # pink
        else:
            color = "\x1b[0m"  # normal
        msg = str(args[1].msg)
        lines = msg.split("\n")

        def color_line(enum_line):
            i, line = enum_line
            tab = " " * (indent - 2) + ": " if i > 0 else ""
            return "%s%s%s%s" % (tab, color, line, "\x1b[0m")

        lines = list(map(color_line, enumerate(lines)))
        args[1].msg = "\n".join(lines)
        return fn(*args)
    # ---
    return new


format_indent = 15
setup_logging_format("%(name)3s|%(levelname)8s : %(message)s")
if DEBUG:
    format_indent = 44
    setup_logging_format("%(name)3s|%(filename)15s:%(lineno)-4s - %(funcName)-15s| %(message)s")

if platform.system() != "Windows":
    emit2 = add_coloring_to_emit_ansi(StreamHandler.emit, format_indent)
    StreamHandler.emit = emit2


def plain(text: str = "", end: str = ""):
    if end == "" and text.rstrip(" ").endswith("\n"):
        text = text.rstrip("\n")
        end = "\n"
    text = "\n".join([
        (" " * 3 + "| " + line)
        for line in text.split('\n')
    ]) + end
    sys.stdout.write(text)
    sys.stdout.flush()


setattr(cpklogger, "print", partial(plain, end="\n"))
setattr(cpklogger, "write", plain)

__all__ = [
    "cpklogger"
]
